// Displays images, loads images from files (png, jpg, etc.), save images, handle color, etc.
// This makes use of the Imaginee++ graphical library, that can be found at
// http://certis.enpc.fr/~keriven/Imagine++/

#ifndef WITH_GRAPHICAL_LIBRARY_H
#define WITH_GRAPHICAL_LIBRARY_H

#include <Imagine/Graphics.h>
using namespace Imagine;

#include "images.h"

namespace MRFSegm {

// A class Image, to contain a 2D-table of int, double or else

typedef RecImage<Color> ColorImage; // Color Image
typedef RecImage<byte> ByteImage;   // 256 grey levels image



// Displays an int-image (grey-scale) or a byte-image.

template <class precis>
void RecImage<precis>::display(double zoom, int p, int q){
  int nxx = int(nx*zoom);
  int nyy = int(ny*zoom);

  byte* tabb = new byte[nxx*nyy];
  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++)
      tabb[i+j*nxx] = byte(limite(0,(*this)(int(i/zoom),int(j/zoom)),255));

  putGreyImage(p, q, tabb, nxx, nyy);

  delete[] tabb;
}

// Displays a color image
template<>
void ColorImage::display(double zoom, int p, int q){

  if (zoom == 1){
    putColorImage(p, q, val, nx, ny);
  } else {
    
    int nxx = int(nx*zoom);
    int nyy = int(ny*zoom);
    
    Color* tabb = new Color[nxx*nyy];
    for (int i = 0; i < nxx; i++)
      for (int j = 0; j < nyy; j++){
	tabb[i+j*nxx].r() = byte(limite(0,(*this)(int(i/zoom),int(j/zoom)).r(),255));
	tabb[i+j*nxx].g() = byte(limite(0,(*this)(int(i/zoom),int(j/zoom)).g(),255));
	tabb[i+j*nxx].b() = byte(limite(0,(*this)(int(i/zoom),int(j/zoom)).b(),255));
      }

    putColorImage(p, q, tabb, nxx, nyy);
    
    delete[] tabb;
  }
}




// Displays a double-image.
// First rescales the intensity.
template <>
void DoubleImage::display(double zoom, int p, int q){
  DoubleImage& A = (* this);
  int nxx = int(A.nx*zoom);
  int nyy = int(A.ny*zoom);

  double zmin = A(0,0);
  double zmax = zmin;
  double z; 

  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++){
      z = A(int(i/zoom),int(j/zoom));
      if (z>zmax) zmax = z;
      if (z < zmin) zmin = z;
    }
  if (zmax == zmin) zmax++;

  byte* tabb = new byte[nxx*nyy];
  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++)
      tabb[i+j*nxx] = byte(255 * ( A(int(i/zoom),int(j/zoom)) - zmin ) / (zmax - zmin) );
  
  putGreyImage(p, q, tabb, nxx, nyy);

  delete[] tabb;
}


// converts from DoubleImage to ByteImage by scaling as much as possible
ByteImage convertit_ByteImage(const DoubleImage& A, double zoom = 1){
  int nxx = int(A.nx*zoom);
  int nyy = int(A.ny*zoom);

  double zmin = A(0,0);
  double zmax = zmin;
  double z; 

  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++){
      z = A(int(i/zoom),int(j/zoom));
      if (z>zmax) zmax = z;
      if (z < zmin) zmin = z;
    }
  if (zmax == zmin) zmax++;

  ByteImage B(nxx,nyy);
  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++)
      B(i,j) = byte(255 * ( A(int(i/zoom),int(j/zoom)) - zmin ) / (zmax - zmin) );
  
  return B;
}






// Constructor int-image loading an image png (or jpg, etc).
// Put full path + name in 'fich'
// Possible zoom factor
template <>
void IntImage::load(const char* fich, double zoom){

  byte * g; int w, h;
  string s;
  
  s = string(fich);

  cout << "Loading the image " << s.c_str() << endl;
  loadGreyImage(s.c_str(), g, w, h);
  cout << "Size before zoom: " << w << " x " << h << endl;

  int nxx = int(w*zoom);
  int nyy = int(h*zoom);
  cout << "Size after zoom: " << nxx << " x " << nyy << endl;
  if ((nx != 0)||(ny != 0))
    empty();
  allocate(nxx,nyy);
  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++)
      (*this)(i,j) = int(g[int(i/zoom)+int(j/zoom)*w]);
  delete[] g;
}



// constructor by loading an image from a (png, jpg) file : requires the definition of 'load' with the graphical library
template <>
IntImage::RecImage(const char * fich, double zoom){
  nx = 0; ny = 0;
  load(fich, zoom);
}


template <>
void DoubleImage::load(const char* fich, double zoom){
  cout << "Warning : loading from a file an integer-valued image to a double-valued image..." << endl;
  empty();
  IntImage A(0,0);
  A.load(fich, zoom);
  copy(A, *this);
  A.empty();
}


template <>
DoubleImage::RecImage(const char * fich, double zoom){
  nx = 0; ny = 0;
  load(fich, zoom);
}



// idem for ByteImage

template <>
void ByteImage::load(const char* fich, double zoom){

  byte * g; int w, h;
  string s;
  
  s = string(fich);

  cout << "Loading the image " << s.c_str() << endl;
  loadGreyImage(s.c_str(), g, w, h);
  cout << "Size before zoom: " << w << " x " << h << endl;

  int nxx = int(w*zoom);
  int nyy = int(h*zoom);
  cout << "Size after zoom: " << nxx << " x " << nyy << endl;
  if ((nx != 0)||(ny != 0))
    empty();
  allocate(nxx,nyy);
  for (int i = 0; i < nxx; i++)
    for (int j = 0; j < nyy; j++)
      (*this)(i,j) = g[int(i/zoom)+int(j/zoom)*w];
  delete[] g;
}



// with color

template <>
void ColorImage::load(const char* fich, double zoom){

  if (zoom != 1)
    cout << "Zoom not taken into account, use convert if necessary." << endl;

  loadColorImage(fich, val, nx, ny);
}



template <>
ColorImage::RecImage(const char * fich, double zoom){
  nx = 0; ny = 0;
  load(fich, zoom);
}




// Saves an image (with format jpg/png/etc.)
// Format depends on the extension of the file name 'fich'
// Possible zoom factor
template <>
void IntImage::save(const char * fich, double zoom){
  if ((nx==0)||(ny==0)){
    cout << "Saving image cancelled because the image is empty." << endl;
    return;
  }

  if (zoom != 1)
    cout << "Warning ! So far when saving an image the zoom factor is taken into account only if < 1, and, yet, only very simply." << endl;

  if (zoom>1){
    cout << "Zoom > 1, changed for zoom = 1" << endl;
    zoom = 1;
  }
  
  byte * g; 
  int w = int(nx*zoom);
  int h = int(ny*zoom);

  cout << "Image saved in " << fich << "... ";
  g = new byte[w*h];
  for (int i = 0; i<w; i++)
    for (int j = 0; j<h; j++)
      g[i+w*j] = byte(limite(0,(*this)(int(i/zoom),int(j/zoom)),255));

  saveGreyImage(fich, g, nx, ny);
  delete[] g;
  cout << "ok" << endl;

}


// idem for ByteImage
template <>
void ByteImage::save(const char * fich, double zoom){
  if ((nx==0)||(ny==0)){
    cout << "Saving image cancelled because the image is empty." << endl;
    return;
  }

  if (zoom != 1)
    cout << "Warning ! So far when saving an image the zoom factor is taken into account only if < 1, and, yet, only very simply." << endl;

  if (zoom>1){
    cout << "Zoom > 1, changed for zoom = 1" << endl;
    zoom = 1;
  }
  
  byte * g; 
  int w = int(nx*zoom);
  int h = int(ny*zoom);

  cout << "Image saved in " << fich << "... ";
  g = new byte[w*h];
  for (int i = 0; i<w; i++)
    for (int j = 0; j<h; j++)
      g[i+w*j] = limite(0,(*this)(int(i/zoom),int(j/zoom)),255);

  saveGreyImage(fich, g, nx, ny);
  delete[] g;
  cout << "ok" << endl;

}



// Saves a binary black&white image 

// Saves an image (with format jpg/png/etc.)
// Format depends on the extension of the file name 'fich'
// Possible zoom factor
template <>
void BoolImage::save(const char * fich, double zoom){
  if ((nx==0)||(ny==0)){
    cout << "Saving image cancelled because the image is empty." << endl;
    return;
  }

  if (zoom != 1)
    cout << "Warning ! So far when saving an image the zoom factor is taken into account only if < 1, and, yet, only very simply." << endl;

  if (zoom>1){
    cout << "Zoom > 1, changed for zoom = 1" << endl;
    zoom = 1;
  }
  
  byte * g; 
  int w = int(nx*zoom);
  int h = int(ny*zoom);

  cout << "Image saved in " << fich << "... ";
  g = new byte[w*h];
  for (int i = 0; i<w; i++)
    for (int j = 0; j<h; j++){
      if ( (*this)(int(i/zoom),int(j/zoom)) ) 
	g[i+w*j] = byte( 255 );
      else 
	g[i+w*j] = byte( 0 );
    }

  saveGreyImage(fich, g, nx, ny);
  delete[] g;
  cout << "ok" << endl;

}



// Saves color image 

template <>
void ColorImage::save(const char * fich, double zoom){
  if ((nx==0)||(ny==0)){
    cout << "Saving image cancelled because the image is empty." << endl;
    return;
  }

  if (zoom != 1)
    cout << "If you wished for another size (zoom != 1) then use the convert command." << endl;

  saveColorImage(fich, val, nx, ny);

}


// Fill a colorImage with Black

template <>
void ColorImage::nul(){
  fill(BLACK);
}


} // end of namespace

#endif
